from enum import Enum


class Site(str, Enum):
    pass


class NyaaSite(Site):
    NYAA_SI = 'https://nyaa.si'
    NYAA_EU = 'https://nyaa.eu'
    NYAA_INK = 'https://nyaa.ink'
    NYAA_LAND = 'https://nyaa.land'
    NYAA_DIGITAL = 'https://nyaa.digital'
    NYAA_ISS_ONE = 'https://nyaa.iss.one'


class SukebeiSite(Site):
    SUKEBEI_NYAA = 'https://sukebei.nyaa.si'
    SKB_NYAA_HACGN_EU = 'https://skb-nyaa.hacgn.eu.org'



class MyaoError(Exception):
    """ Base Myao error """


class FormatterError(MyaoError):
    """ Base Formatter error """


class InvalidCategory(FormatterError):
    """ Invalid category error """


class InvalidSubcategory(FormatterError):
    """ Invalid subcategory error """


class InvalidPage(FormatterError):
    """ Invalid page error """


class ExtractionError(MyaoError):
    """ Base Extraction error """


class NoResultsFound(ExtractionError):
    """ No results found error """


class UnexpectedError(ExtractionError):
    """ Unexpected error """
    def __init__(self, exc: BaseException, error_msg: str):
        self.exc = exc
        super().__init__(error_msg)



from .containers import (
    Torrent,
    Comment,
    View,
    User
)

from .sites import (
    Site,
    NyaaSite,
    SukebeiSite
)

from .extractors import (
    Extractor,
    Parser,
    get_multiple_torrents,
    get_multiple_torrents_rss,
    get_single_torrent,
    get_description,
    get_comments,
    get_view
)

from .parameters import (
    Category,
    Subcategory,
    Order,
    Sorting,
    Filter
)

from .errors import (
    MyaoError,
    FormatterError,
    InvalidCategory,
    InvalidPage,
    InvalidSubcategory,
    ExtractionError,
    NoResultsFound,
)

from .urls import (
    Formatter,
    format_url,
    get_user_url,
    get_single_torrent_url,
    get_download_torrent_url,
)

from .magnets import magnet_builder


__all__ = [
    # Containers
    'Torrent',
    'Comment',
    'Comment',
    'View',
    'User',

    # Extractors
    'Extractor',
    'Parser',
    'get_multiple_torrents',
    'get_multiple_torrents_rss',
    'get_single_torrent',
    'get_description',
    'get_comments',
    'get_view',

    # Sites
    'Site',
    'NyaaSite',
    'SukebeiSite',

    # Parameters
    'Category',
    'Subcategory',
    'Order',
    'Sorting',
    'Filter',

    # Urls
    'Formatter',
    'format_url',
    'get_user_url',
    'get_single_torrent_url',
    'get_download_torrent_url',

    # Errors
    'MyaoError',
    'FormatterError',
    'InvalidCategory',
    'InvalidSubcategory',
    'InvalidPage',
    'ExtractionError',
    'NoResultsFound',

    # Magnet
    'magnet_builder'
]


__version__ = "0.0.2"



from datetime import datetime
from dataclasses import dataclass
from typing import List, Optional
from myao.parameters import Category, Subcategory, TorrentType


@dataclass(unsafe_hash=True)
class User:
    name: str
    trusted: bool


@dataclass(unsafe_hash=True)
class Torrent:
    size: int
    date: datetime
    seeders: int
    leechers: int
    completed_downloads: int
    category: Category
    subcategory: Subcategory
    title: str
    code: int
    magnet_link: str
    comments: int = 0
    torrent_type: Optional[TorrentType] = None
    information: Optional[str] = None
    info_hash: Optional[str] = None
    submitter: Optional[User] = None


@dataclass(unsafe_hash=True)
class Comment:
    user: User
    date: str
    text: str
    avatar: Optional[str]


@dataclass(unsafe_hash=True)
class View:
    torrent: Torrent
    description: str
    comments: List[Comment]



_UNITS = {
    'KiB': 1024, 
    'MiB': 1024**2,
    'GiB': 1024**3,
    'TiB': 1024**4
}


def convert_to_bytes(size: str) -> int:
    num, unit = size.split()
    return int(float(num) * _UNITS[unit])



from dataclasses import dataclass
from enum import Enum
from typing import Dict, Set


class Subcategory(Enum):
    ANIME_MUSIC_VIDEO = 'Anime Music Video'              # Nyaa
    ENGLISH_TRANSLATED = 'English-translated'            # Nyaa
    NON_ENGLISH_TRANSLATED = 'Non-English-translated'    # Nyaa
    RAW = 'Raw'                                          # Nyaa
    LOSSLESS = 'Lossless'                                # Nyaa
    LOSSY = 'Lossy'                                      # Nyaa
    IDOL_PROMOTIONAL_VIDEO = 'Idol/Promotional Video'    # Nyaa
    GRAPHICS = 'Graphics'                                # Nyaa
    PHOTOS = 'Photos'                                    # Nyaa
    APPLICATIONS = 'Applications'                        # Nyaa
    GAMES = 'Games'                                      # Nyaa - Sukebei
    ANIME = 'Anime'                                      # Sukebei
    MANGA = 'Manga'                                      # Sukebei
    DOUJINSHI = 'Doujinshi'                              # Sukebei
    PICTURES = 'Pictures'                                # Sukebei
    PHOTOBOOKS_AND_PICTURES = 'Photobooks / Pictures'    # Sukebei
    VIDEOS = 'Videos'                                    # Sukebei


@dataclass
class _Category:
    name: str
    param: int
    subcategories: Dict[Subcategory, int]


class Category(Enum):
    ALL_CATEGORIES = _Category(name='All categories', param=0, subcategories={})

    # Nyaa
    ANIME = _Category(name='Anime', param=1, subcategories={
        Subcategory.ANIME_MUSIC_VIDEO: 1,
        Subcategory.ENGLISH_TRANSLATED: 2,
        Subcategory.NON_ENGLISH_TRANSLATED: 3,
        Subcategory.RAW: 4,
    })
    AUDIO = _Category(name="Audio", param=2, subcategories={
        Subcategory.LOSSLESS: 1,
        Subcategory.LOSSY: 2,
    })
    LITERATURE = _Category(name='Literature', param=3, subcategories={
        Subcategory.ENGLISH_TRANSLATED: 1,
        Subcategory.NON_ENGLISH_TRANSLATED: 2,
        Subcategory.RAW: 3,
    })
    LIVE_ACTION = _Category(name='Live Action', param=4, subcategories={
        Subcategory.ENGLISH_TRANSLATED: 1,
        Subcategory.IDOL_PROMOTIONAL_VIDEO: 2,
        Subcategory.NON_ENGLISH_TRANSLATED: 3,
        Subcategory.RAW: 4,
    })
    PICTURES = _Category(name='Pictures', param=5, subcategories={
        Subcategory.GRAPHICS: 1,
        Subcategory.PHOTOS: 2,
    })
    SOFTWARE = _Category(name='Software', param=6, subcategories={
        Subcategory.APPLICATIONS: 1,
        Subcategory.GAMES: 2,
    })

    # Sukebei
    ART = _Category(name='Art', param=1, subcategories={
        Subcategory.ANIME: 1,
        Subcategory.DOUJINSHI: 2,
        Subcategory.GAMES: 3,
        Subcategory.MANGA: 4,
        Subcategory.PICTURES: 5,
    })
    REAL_LIFE = _Category(name='Real Life', param=2, subcategories={
        Subcategory.PHOTOBOOKS_AND_PICTURES: 1,
        Subcategory.VIDEOS: 2,
    })

    @classmethod
    def get_category_by_name(cls, category_name: str) -> 'Category':
        return Category.__members__[category_name.upper().replace(' ', '_')]

    @classmethod
    def get_nyaa_categories(cls) -> Set['Category']:
        return {category for category in Category if category.is_nyaa_category()}

    @classmethod
    def get_subekei_categories(cls) -> Set['Category']:
        return {category for category in Category if category.is_sukebei_category()}

    def is_sukebei_category(self) -> bool:
        return self == Category.ALL_CATEGORIES or self in (Category.REAL_LIFE, Category.ART)

    def is_nyaa_category(self) -> bool:
        return self == Category.ALL_CATEGORIES or not self.is_sukebei_category()


class Sorting(Enum):
    SEEDERS = 'seeders'
    LEECHERS = 'leechers'
    ID = 'id'
    DOWNLOADS = 'downloads'
    COMMENTS = 'comments'


class Order(Enum):
    ASCENDING = 'asc'
    DESCENDANT = 'desc'


@dataclass
class _Filter:
    name: str
    param: int


class Filter(Enum):
    NO_FILTER = _Filter(name='No filter', param=0)
    NO_REMAKES = _Filter(name='No remakes', param=1)
    TRUSTED_ONLY = _Filter(name='Trusted only', param=2)


class TorrentType(Enum):
    REMAKE = "remake"
    TRUSTED = "trusted"
    DEFAULT = "default"



import logging
from enum import Enum
from datetime import datetime
from typing import Callable, List, Union
from bs4 import BeautifulSoup, FeatureNotFound
from myao.containers import Comment, Torrent, User, View
from myao.parameters import Category, Subcategory, TorrentType
from myao.errors import NoResultsFound, UnexpectedError
from myao.magnets import magnet_builder
from myao.utils import convert_to_bytes


logger = logging.getLogger(__name__)


class Parser(Enum):
    HTML = "html.parser"
    LXML = "lxml"
    HTML5LIB = "html5lib"


class Extractor:
    def __init__(self, parser: Parser = Parser.HTML):
        self.parser = parser

    def get_multiple_torrents(self, content: Union[str, bytes]) -> List[Torrent]:
        return get_multiple_torrents(content, self.parser)

    def get_multiple_torrents_rss(self, content: Union[str, bytes]) -> List[Torrent]:
        return get_multiple_torrents_rss(content, self.parser)

    def get_view(self, content: Union[str, bytes]) -> View:
        return get_view(content, self.parser)

    def get_single_torrent(self, content: Union[str, bytes]) -> Torrent:
        return get_single_torrent(content, self.parser)

    def get_comments(self, content: Union[str, bytes]) -> List[Comment]:
        return get_comments(content, self.parser)

    def get_description(self, content: Union[str, bytes]) -> str:
        return get_description(content, self.parser)


def _log_unexpected_errors(func: Callable) -> Callable:
    def _inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (NoResultsFound, FeatureNotFound)as e:
            raise e
        except Exception as e:
            msg = f"Please provide this log error to https://github.com/g3nsy/myao/issues"
            logger.error(e)
            logger.error(msg)
            raise UnexpectedError(e, msg)
    return _inner


@_log_unexpected_errors
def get_multiple_torrents(content: Union[str, bytes], parser=Parser.HTML) -> List[Torrent]:
    soup = BeautifulSoup(content, parser.value)
    tbody_entries = soup.find('tbody')
    if not tbody_entries:
        raise NoResultsFound
    torrents: List[Torrent] = []
    for entry in tbody_entries.find_all('tr'):  # type: ignore

        cols = entry.find_all('td')
        col0 = cols[0].find_all('a')
        col1 = cols[1].find_all('a')
        col2 = cols[2].find_all('a')

        category_name, subcategory_name = col0[0]['title'].split(' - ')
        category = Category.get_category_by_name(category_name)
        subcategory = Subcategory(subcategory_name)

        class_value = entry.get('class')
        torrent_type = (
            TorrentType.TRUSTED if class_value == "success"
            else TorrentType.REMAKE if class_value == "danger"
            else TorrentType.DEFAULT
        )

        if len(col1) == 2:
            title = col1[1]['title']
            comments = int(col1[0]['title'].split()[0])
        else:
            title = col1[0]['title']
            comments = 0

        seeders = int(cols[5].get_text())
        leechers = int(cols[6].get_text())
        completed_downloads = int(cols[7].get_text())
        date = datetime.strptime(cols[4].get_text(), '%Y-%m-%d %H:%M')
        size = convert_to_bytes(cols[3].get_text())
        magnet_link = col2[1]['href']
        code = int(col2[0]['href'][10:-8])

        torrents.append(
            Torrent(
                title=title,
                category=category,
                subcategory=subcategory,
                seeders=seeders,
                leechers=leechers,
                comments=comments,
                completed_downloads=completed_downloads,
                date=date,
                size=size,
                torrent_type=torrent_type,
                magnet_link=magnet_link,
                code=code,
            )
        )
    return torrents


@_log_unexpected_errors
def get_multiple_torrents_rss(content: Union[str, bytes], parser=Parser.HTML) -> List[Torrent]:
    soup = BeautifulSoup(content, parser.value)
    torrents: List[Torrent] = []
    items = soup.find_all('item')
    if not items:
        raise NoResultsFound
    for item in items:
        category_name, subcategory_name = item.find('nyaa:category').get_text().split(' - ')
        category = Category.get_category_by_name(category_name.strip())
        subcategory = Subcategory(subcategory_name.strip())

        is_trusted = item.find('nyaa:trusted') == 'Yes'
        is_remake = item.find('nyaa:remake') == 'Yes'

        torrent_type = (
            TorrentType.TRUSTED if is_trusted
            else TorrentType.REMAKE if is_remake
            else TorrentType.DEFAULT
        )

        title = item.title.text
        info_hash = item.find('nyaa:infohash').text
        seeders = int(item.find('nyaa:seeders').text)
        leechers = int(item.find('nyaa:leechers').text)
        comments = int(item.find('nyaa:comments').text)
        completed_downloads = int(item.find('nyaa:downloads').text)
        date = datetime.strptime(item.find('pubdate').text, '%a, %d %b %Y %H:%M:%S %z')
        size = convert_to_bytes(item.find('nyaa:size').text)
        code = int(item.find('guid').get_text().split('/')[-1])
        magnet_link = magnet_builder(info_hash, title)

        torrents.append(
            Torrent(
                title=title,
                category=category,
                subcategory=subcategory,
                seeders=seeders,
                leechers=leechers,
                comments=comments,
                completed_downloads=completed_downloads,
                date=date,
                size=size,
                torrent_type=torrent_type,
                info_hash=info_hash,
                magnet_link=magnet_link,
                code=code,
            )
        )
    return torrents


@_log_unexpected_errors
def get_view(content: Union[str, bytes], parser=Parser.HTML) -> View:
    soup = BeautifulSoup(content, parser.value)
    torrent = _get_single_torrent(soup)
    comments = _get_comments(soup)
    description = _get_description(soup)
    return View(
        torrent=torrent,
        description=description,
        comments=comments
    )


@_log_unexpected_errors
def get_description(content: Union[str, bytes], parser=Parser.HTML) -> str:
    return _get_description(BeautifulSoup(content, parser.value))


def _get_description(soup: BeautifulSoup) -> str:
    return soup.find('div', {'id': 'torrent-description'}).get_text()  # type: ignore


@_log_unexpected_errors
def get_comments(content: Union[str, bytes], parser=Parser.HTML) -> List[Comment]:
    return _get_comments(BeautifulSoup(content, parser.value))


def _get_comments(soup: BeautifulSoup) -> List[Comment]:
    comments_data = soup.find('div', {'id': 'comments'})
    comment_class = {'class': 'panel panel-default comment-panel'}
    comments: List[Comment] = []
    for comment_data in comments_data.find_all('div', comment_class):  # type: ignore
        try:
            user = User(
                name=comment_data.find('a', {'class': 'text-default'}).get_text(),
                trusted=False
            )
        except AttributeError:
            user = User(
                name=comment_data.find('a', {'class': 'text-success'}).get_text(),
                trusted=True
            )

        date = comment_data.find('small').get_text()
        text = comment_data.find('div', {'class': 'comment-content'}).get_text()

        avatar = comment_data.find('img')['src']
        if avatar.startswith('/'):
            avatar = None

        comments.append(Comment(
            user=user,
            avatar=avatar,
            date=date,
            text=text
        ))

    return comments


@_log_unexpected_errors
def get_single_torrent(content: Union[str, bytes], parser=Parser.HTML) -> Torrent:
    return _get_single_torrent(BeautifulSoup(content, parser.value))


def _get_single_torrent(soup: BeautifulSoup) -> Torrent:
    title = soup.find('h3', {'class': 'panel-title'}).get_text(strip=True)  # type: ignore
    panel_body = soup.find('div', {'class': 'panel-body'})
    rows = panel_body.find_all('div', {'class': 'row'})  # type: ignore

    comments_data = soup.find(
        'div', {'id': 'comments'}
    ).find('h3', {'class': 'panel-title'})  # type: ignore
    comments = int(comments_data.get_text().split(' - ')[1])  # type: ignore

    div_row_0 = rows[0].find_all('div')
    div_row_1 = rows[1].find_all('div')
    div_row_2 = rows[2].find_all('div')
    div_row_3 = rows[3].find_all('div')
    div_row_4 = rows[4].find_all('div')

    category_name, subcategory_name = div_row_0[1].get_text().split(' - ')
    category = Category.get_category_by_name(category_name.strip())
    subcategory = Subcategory(subcategory_name.strip())

    date = datetime.strptime(div_row_0[3].get_text(strip=True), '%Y-%m-%d %H:%M %Z')

    try:
        trusted = div_row_1[1].find('a')['title'] == 'Trusted'
    except TypeError:  # Anonymous
        trusted = False

    name = div_row_1[1].get_text(strip=True)
    submitter = User(name=name, trusted=trusted)
    seeders = int(div_row_1[3].get_text())
    information = div_row_2[1].get_text()
    leechers = int(div_row_2[3].get_text())
    size = convert_to_bytes(div_row_3[1].get_text())
    completed_downloads = int(div_row_3[3].get_text())
    info_hash = div_row_4[1].get_text()
    footer_attrs = soup.find(
        'div', {'class': 'panel-footer clearfix'}
    ).find_all('a')  # type: ignore
    code = int(footer_attrs[0]['href'][10:-8])
    magnet_link = footer_attrs[1]['href']
    information = information if information != 'No information.' else None

    return Torrent(
        size=size,
        date=date,
        category=category,
        subcategory=subcategory,
        submitter=submitter,
        completed_downloads=completed_downloads,
        code=code,
        magnet_link=magnet_link,
        seeders=seeders,
        leechers=leechers,
        title=title,
        information=information,
        info_hash=info_hash,
        comments=comments
    )



from urllib.parse import urlencode, quote


_KNOWN_TRACKERS = [
    "http://nyaa.tracker.wf:7777/announce",
    "udp://open.stealth.si:80/announce",
    "udp://tracker.opentrackr.org:1337/announce",
    "udp://exodus.desync.com:6969/announce",
    "udp://tracker.torrent.eu.org:451/announce",
]


def magnet_builder(info_hash: str, title: str) -> str:
    magnet_link = (
        f"magnet:?xt=urn:btih:{info_hash}&" 
        + urlencode({"dn": title}, quote_via=quote)
    )
    for tracker in _KNOWN_TRACKERS:
        magnet_link += f"&{urlencode({'tr': tracker})}"
    return magnet_link



from typing import Optional, Union
from myao.sites import Site, NyaaSite, SukebeiSite
from myao.errors import (
    InvalidCategory,
    InvalidSubcategory,
    InvalidPage
)
from myao.parameters import (
    Category,
    Subcategory,
    Filter,
    Order,
    Sorting
)


class Formatter:
    def __init__(self):
        self._site: Site = NyaaSite.NYAA_SI
        self._user: Optional[str] = None
        self._query = ''
        self._page = 0
        self._category = Category.ALL_CATEGORIES
        self._subcategory: Optional[Subcategory] = None
        self._sorting = Sorting.ID
        self._order = Order.DESCENDANT
        self._filter = Filter.NO_FILTER
        self._rss = False

    @property
    def site(self) -> Site:
        return self._site

    @site.setter
    def site(self, site: Site) -> None:
        if self._site != site:
            self._site = site
            self._category = Category.ALL_CATEGORIES
            self._subcategory = None

    @property
    def user(self) -> Optional[str]:
        return self._user

    @user.setter
    def user(self, user: str) -> None:
        self._user = user

    @property
    def query(self) -> Optional[str]:
        return self._query

    @query.setter
    def query(self, query: str = '') -> None:
        self._query = '+'.join(query.split())

    @property
    def page(self) -> int:
        return self._page

    @page.setter
    def page(self, page: int) -> None:
        if page < 0:
            raise InvalidPage(page)
        else:
            self._page = page

    @property
    def category(self) -> Category:
        return self._category

    @category.setter
    def category(self, category: Optional[Category]) -> None:
        if not category or category is Category.ALL_CATEGORIES:
            self._category = Category.ALL_CATEGORIES
            self._subcategory = None
        else:
            if self._category != category:
                if (
                    isinstance(self._site, NyaaSite) and category.is_sukebei_category()
                    or isinstance(self._site, SukebeiSite) and category.is_nyaa_category()
                ):
                    raise InvalidCategory(
                        f'Invalid category for site {self._site.value}: {category}'
                    )
                else:
                    self._category = category
                    self._subcategory = None

    @property
    def subcategory(self) -> Optional[Subcategory]:
        return self._subcategory

    @subcategory.setter
    def subcategory(self, subcategory: Optional[Subcategory]) -> None:
        if subcategory and subcategory not in self._category.value.subcategories:
            raise InvalidSubcategory(
                'Invalid subcategory for category '
                f'"{self._category.value.name}": {subcategory.name}'
            )
        self._subcategory = subcategory

    @property
    def sorting(self) -> Sorting:
        return self._sorting

    @sorting.setter
    def sorting(self, sorting: Optional[Sorting]) -> None:
        if not sorting:
            self._sorting = Sorting.ID
        else:
            self._sorting = sorting

    @property
    def order(self) -> Order:
        return self._order

    @order.setter
    def order(self, order: Optional[Order]) -> None:
        if not order:
            self._order = Order.DESCENDANT
        else:
            self._order = order

    @property
    def filter(self) -> Filter:
        return self._filter

    @filter.setter
    def filter(self, filter_: Optional[Filter]) -> None:
        if not filter_:
            self._filter = Filter.NO_FILTER
        else:
            self._filter = filter_

    @property
    def rss(self) -> bool:
        return self._rss

    @rss.setter
    def rss(self, rss: bool) -> None:
        self._rss = rss

    def format(self) -> str:
        return '{}/?q={}&p={}&c={}_{}&f={}&s={}&o={}&u={}&page={}'.format(
            self.site.value,
            self.query,
            self.page,
            self.category.value.param,
            self.category.value.subcategories[self.subcategory] if self.subcategory else 0,
            self.filter.value.param,
            self.sorting.value,
            self.order.value,
            self.user if self.user else '',
            'rss' if self.rss else ''
        )

    def get_download_torrent_url(self, code: Union[str, int]) -> str:
        return get_download_torrent_url(code, site=self.site)

    def get_single_torrent_url(self, code: Union[str, int]) -> str:
        return get_single_torrent_url(code, site=self.site)

    def get_user_url(self, user: str) -> str:
        return get_user_url(user, site=self.site, rss=self.rss)


def format_url(
    query='',
    *,
    user='',
    site: Site = NyaaSite.NYAA_SI,
    page=0,
    category: Optional[Category] = Category.ALL_CATEGORIES,
    subcategory: Optional[Subcategory] = None,
    filter_: Optional[Filter] = Filter.NO_FILTER,
    sorting: Optional[Sorting] = Sorting.ID,
    order: Optional[Order] = Order.DESCENDANT,
    rss=False,
) -> str:
    formatter = Formatter()
    formatter.query = query
    formatter.site = site
    formatter.user = user
    formatter.page = page
    formatter.category = category
    formatter.subcategory = subcategory
    formatter.filter = filter_
    formatter.sorting = sorting
    formatter.order = order
    formatter.rss = rss
    return formatter.format()


def get_download_torrent_url(code: Union[str, int], *, site: Site = NyaaSite.NYAA_SI) -> str:
    return f"{site.value}/download/{code}.torrent"


def get_single_torrent_url(code: Union[str, int], *, site: Site = NyaaSite.NYAA_SI) -> str:
    return f"{site.value}/view/{code}"


def get_user_url(user: str, *, site: Site = NyaaSite.NYAA_SI, rss=False) -> str:
    return f'{site.value}/?u={user}{"&page=rss" if rss else ""}'
